
#include <stdio.h>
#include <fpe.h>


void reg_to_double(fpe_register *a,uint32 *b)
{
#ifdef LITTLE_ENDIAN
	int exp;
	
	b[1]=(a->sign<<31);
	exp=a->exp-0x3fff+0x3ff;
	if(exp<0){
		/* underflow */
		return;
	}
	if(exp>0x7ff){
		/* overflow */
		return;
	}
	b[1]|=exp<<20;
	
	b[1]|=0xfffff&(a->man_hi>>11);
	b[0]=a->man_hi<<21;
	b[0]|=a->man_lo>>11;
#else
#endif
}

void double_to_reg(fpe_register *a,uint32 *b)
{
#ifdef LITTLE_ENDIAN
	a->sign=b[1]>>31;
	
	a->exp=(b[1]>>20)-0x3ff+0x3fff;
	
	a->man_hi=0x80000000 | (b[1]<<11) | (b[0]>>21);
	a->man_lo=(b[0]<<11);

#else
#endif
}

void dump_register(fpe_register *a)
{
	printf("sign=%c [%d]\n",a->sign?'-':'+',a->sign);
	printf("exponent=%d [0x%04x]\n",a->exp-0x3fff,a->exp);
	printf("mantissa=0x%08lx%08lx\n",a->man_hi,a->man_lo);
}

void dump_result(fpe_result *a)
{
	printf("sign=%c [%d]\n",a->sign?'-':'+',a->sign);
	printf("exponent=%d [0x%04x]\n",a->exp-0x3fff,a->exp);
	printf("overflow=0x%08lx\n",a->man_overflow);
	printf("mantissa=0x%08lx%08lx\n",a->man_hi,a->man_lo);
	printf("underflow=0x%08lx\n",a->man_underflow);
}

void dump_double(double a)
{
#ifdef LITTLE_ENDIAN
	printf("0x%08lx%08lx\n",((uint32 *)&a)[1],((uint32 *)&a)[0]);
#else
#endif
}

