
#include <fpe.h>

void mul(fpe_register *src,fpe_register *dest)
{
	uint32 a[5];
	uint32 b[5];
	uint32 c[9];
	uint32 tmp;
	int i,j;

	result->exp = src->exp + dest->exp - 0x3fff + 1;
	result->sign = src->sign ^ dest->sign;
	
	a[1]=src->man_hi>>16;
	a[2]=src->man_hi&0xffff;
	a[3]=src->man_lo>>16;
	a[4]=src->man_lo&0xffff;
	
	b[1]=dest->man_hi>>16;
	b[2]=dest->man_hi&0xffff;
	b[3]=dest->man_lo>>16;
	b[4]=dest->man_lo&0xffff;
	
	for(i=0;i<9;i++)c[i]=0;
	
	for(i=1;i<=4;i++)for(j=1;j<=4;j++){
		tmp=a[i]*b[j];
		c[i+j]+=tmp&0xffff;
		c[i+j-1]+=tmp>>16;
	}
	for(i=7;i>=0;i--){
		c[i]+=c[i+1]>>16;
	}
	result->man_overflow=c[0];
	result->man_hi=(c[1]<<16) | (c[2]&0xffff);
	result->man_lo=(c[3]<<16) | (c[4]&0xffff);
	result->man_underflow=(c[5]<<16) | (c[6]&0xffff);
}

/* this is for adding numbers that have the same sign
   or subtracting numbers with different sign.
   result takes sign of src */

void add(fpe_register *src,fpe_register *dest)
{
	uint32 src_hi,src_lo;
	uint32 c1,c2,c3,c4;
	int bits;
	
	result->sign=src->sign;
	src_hi=src->man_hi;
	src_lo=src->man_lo;

	bits=src->exp-dest->exp;
	if(bits>0){
		result->exp=src->exp;
		mantissa_ror(bits,&dest->man_hi,&dest->man_lo);
	}else{
		result->exp=dest->exp;
		mantissa_ror(-bits,&src_hi,&src_lo);
	}
	
	c4=(dest->man_lo&0xffff)+(src_lo&0xffff);
	c3=(dest->man_lo>>16)+(src_lo>>16)+(c4>>16);
	c2=(dest->man_hi&0xffff)+(src_hi&0xffff)+(c3>>16);
	c1=(dest->man_hi>>16)+(src_hi>>16)+(c2>>16);
	
	result->man_overflow=(c1>>16);
	result->man_hi=(c1<<16) | (c2&0xffff);
	result->man_lo=(c3<<16) | (c4&0xffff);
}

/* see note for add() */

void sub(fpe_register *src,fpe_register *dest)
{
	uint32 pos_hi,pos_lo;
	uint32 neg_hi,neg_lo;
	uint32 c1,c2,c3,c4;
	
	int bits=src->exp-dest->exp;
	if(bits>0){
		result->sign=src->sign;
		result->exp=src->exp;
		pos_hi=src->man_hi;
		pos_lo=src->man_lo;
		neg_hi=dest->man_hi;
		neg_lo=dest->man_lo;
	}else{
		result->sign=dest->sign;
		result->exp=dest->exp;
		pos_hi=dest->man_hi;
		pos_lo=dest->man_lo;
		neg_hi=src->man_hi;
		neg_lo=src->man_lo;
	}
	mantissa_ror(-bits,&neg_hi,&neg_lo);
	
	c4=(pos_lo&0xffff) - (neg_lo&0xffff);
	c3=(pos_lo>>16)    - (neg_lo>>16)    - (c4>=0x10000);
	c2=(pos_hi&0xffff) - (neg_hi&0xffff) - (c3>=0x10000);
	c1=(pos_hi>>16)    - (neg_hi>>16)    - (c2>=0x10000);
	
	result->man_hi=(c1<<16) | (c2&0xffff);
	result->man_lo=(c3<<16) | (c4&0xffff);
	if(c1>=0x10000){
		result->sign^=1;
		result->man_hi=~result->man_hi;
		result->man_lo=~result->man_lo;
	}
	result->man_overflow=0;
}

void mantissa_ror(unsigned bits,uint32 *hi,uint32 *lo)
{
	if(bits<1)return;
	if(bits<32){
		(*lo)>>=bits;
		(*lo)|=((*hi)<<(32-bits));
		(*hi)>>=bits;
		return;
	}
	if(bits<64){
		(*lo)=(*hi)>>(bits-32);
		(*hi)=0;
		return;
	}
	*lo=0;
	*hi=0;
}

void mantissa_rol(unsigned bits,uint32 *hi,uint32 *lo)
{
	if(bits<1)return;
	if(bits<32){
		(*hi)<<=bits;
		(*hi)|=((*lo)>>(32-bits));
		(*lo)<<=bits;
		return;
	}
	if(bits<64){
		(*hi)=(*lo)<<(bits-32);
		(*lo)=0;
		return;
	}
	*lo=0;
	*hi=0;
}

void normalize(fpe_register *dest)
{
	uint32 tmp;
	unsigned bits;
	
	dest->sign=result->sign;
	
	if(result->man_overflow>0){
		bits=0;
		tmp=result->man_overflow;
		while(tmp){bits++;tmp>>=1;}
		mantissa_ror(bits,&result->man_hi,&result->man_lo);
		result->man_hi|=(result->man_overflow<<(32-bits));
		result->exp+=bits;
	}else if((result->man_hi&0x8000000)==0){
		if(result->man_hi==0 && result->man_lo==0){
			result->exp=0;
		}else{
			if(result->man_hi==0){
				result->exp-=32;
				result->man_hi=result->man_lo;
				result->man_lo=result->man_underflow;
				result->man_underflow=0;
			}
			bits=0;
			tmp=result->man_hi;
			while(!(tmp&0x80000000)){bits++;tmp<<=1;}
			mantissa_rol(bits,&result->man_hi,&result->man_lo);
			result->man_lo|=(result->man_underflow>>(32-bits));
			result->exp-=bits;
		}
	}
	dest->exp=result->exp;
	dest->man_hi=result->man_hi;
	dest->man_lo=result->man_lo;
}

