|
|  fp_fadd: Perform an FADD operation.
|

fp_fadd:
	|
	|  First, some preliminaries...
	|
	and.l	#0x00FF00FF, (fpe_fpsr,%a4)  | clear FPSR bytes that will be changed
	move.l	(%a0), %d0		| load source operand into %d0-%d2
	move.l	(#4,%a0), %d1
	move.l	(#8,%a0), %d2
	move.l	(%a1), %d4		| load destination operand into %d4-%d6
	move.l	(#4,%a1), %d5
	move.l	(#8,%a1), %d6
	jsr	fpd_handle_nan		| handle Not-A-Number inputs, if any

	|
	|  Special case handling - bear with me here
	|
	bfextu	%d0{#16,#15}, %d7	| is source an infinity?
	cmpi.w	#0x7FFF, %d7		
	blt	idINF
	bfextu	%d4{#16,#15}, %d7	| source is INF... is dest an INF?
	cmpi.w	#0x7FFF, %d7
	bne	fpd_gen_inf		| if dest isn't INF, return inf w/sign of source
	xor.l	%d0, %d4		| so dest is INF... do the signs differ?
	btst	#31, %d4
	bne	fpd_gen_nan		| if so, return NAN
	bra	fpd_gen_inf		| if not, return infinity
idINF:	bfextu	%d4{#16,#15}, %d7	| is dest an infinity?
	cmpi.w	#0x7FFF, %d7
	blt	isZRO
	move.l	%d4, %d0
	bra	fpd_gen_inf		| yes... so leave it untouched (effectively)
isZRO:	move.l	%d1, %d7		| is source a zero?
	or.l	%d2, %d7
	bne 	idZRO
	move.l	%d5, %d7		| source is zero. is dest zero of different sign?
	or.l	%d6, %d7
	beq	diZRO
	move.l	%d4, %d0
	move.l	%d5, %d1
	move.l	%d6, %d2
	bra	fpd_round		| dest isn't zero, so leave it be (effectively)
diZRO:	move.l	%d0, %d7		| dest is zero... does its sign differ?
	xor.l	%d4, %d7
	bftst	#31, %d7
	bne	sdZRO
	bra	fpd_gen_zero		| dest is zero of same sign... leave it be
sdZRO:	clr.l	%d0			| signs differ... see if rounding mode is RM
	bfextu	(fpe_fpcr,%a4){#4,#2}, %d7
	subq.l	#2, %d7			| cheaper than cmpi.b #2, %d7
	bne	1f			| if not RM mode, return +0.0
	bset	#31, %d0		| if RM mode, return -0.0
	bra	fpd_gen_zero
idZRO:	move.l	%d5, %d7		| is dest a zero?
	or.l	%d6, %d7
	beq 	fpd_round		| if so, return source

	|
	|  Normalize operands and put "heavier" (larger-magnitude) operand
	|  into %d0-%d2.  (We want abs(%d0-%d2) > abs(%d4-%d6).)
	|
unify:	jsr	fpd_norm_dyadic
	bfextu	%d0{#16,#15}, %d3
	bfextu	%d4{#16,#15}, %d7
	cmp.w	%d3, %d7
	blt	opswpd
	bgt	swapem
	cmp.l	%d1, %d5
	blt	opswpd
	bgt	swapem
	cmp.l	%d2, %d6
	blt	opswpd
swapem:	exg	%d0, %d4
	exg	%d1, %d5
	exg	%d2, %d6

	|
	|  Check to see if the mismatch between exponents is too large
	|  to make an addition operation meaningful.  If so, either
	|  just return the %d0-%d2 operand, or "bump" the smaller value
	|  up to the smallest denormalized mantissa with exponent equal
	|  to the value in %d0-%d2, depending on sign and rounding mode:
	|
	|  if %d0-%d2 negative, %d4-%d6 negative: bump to +denorm if RM
	|  if %d0-%d2 negative, %d4-%d6 positive: bump to -denorm if RP
	|  if %d0-%d2 positive, %d4-%d6 negative: bump to -denorm if RM
	|  if %d0-%d2 positive, %d4-%d6 positive: bump to +denorm if RP
	|
opswpd:	bfextu	%d0{#16,#15}, %d3	| is exponent mismatch greater than 96 bits?
	bfextu	%d4{#16,#15}, %d7
	sub.w	%d7, %d3
	cmpi.w	#0x60, %d3
	blt	inrng
	or.w	#0x0204, (fpe_fpsr+2, %a4)  | large mismatch -  set INEX2 bit in FPSR
	move.w	(fpe_fpcr+2,%a4), %d3	| check round mode
	btst	#5, %d3			| if RN or RZ, don't bother
	beq	fpd_round
	ror.l	#5, %d3
	eor.l	%d4, %d3
	btst	#31, %d3
	beq	fpd_round
	eor.l	%d0, %d4		| we survived - bump %d4-%d7
	clr.l	%d5			| (we don't mess with exponent, since it's
	moveq.l	#1, %d6			|  presumed equal to that in %d0 from here)
inrng:	clr.l	%d7			| exponent mismatch is in manageable range...
	cmp.w	#0x20, %d3		| shift mantissas so they align
	blt	dsra
	move.l	%d6, %d7
	move.l	%d5, %d6
	clr.l	%d5
	sub.w	#0x20, %d3
	bra	inrng
dsra:	tst.l	%d3			| special case check... identical exponents?
	beq	ops
	move.l	%d0, -(%sp)		| now do fine-grained bit shifting
	movem.l	%d5-%d7, -(%sp)		| (dammit, I'm running out of registers...
	lsr.l	%d3, %d5		|     shoulda saved up for a Power Mac :-)
	lsr.l	%d3, %d6
	lsr.l	%d3, %d7
	neg.w	%d3
	add.w	#32, %d3
	move.l	+(%sp), %d0
	lsl.l	%d3, %d0
	or.l	%d0, %d6
	move.l	+(%sp), %d0
	lsl.l	%d3, %d0
	or.l	%d0, %d7
	move.l	+(%sp), %d0
	lsl.l	%d3, %d0
	beq	1f
	bset	#0, %d7
	move.l	+(%sp), %d0

	|
	|  Now the real fun begins.
	|
ops:	move.l	%d0, %d3		| check if signs differ
	eor.l	%d4, %d3
	btst	31, %d3 
	bne	sub
	move.l	%d7, %d3		| signs are the same, so, finally... add!
	add.l	%d6, %d2
	addx.l	%d5, %d1
	bcc	fpd_round
	roxr.l	#1, %d1
	roxr.l	#1, %d2
	roxr.l	#1, %d3
	bfextu	%d0{#16,#15}, %d7
	addq.w	#1, %d7
	bfins	%d7, %d0{#16,#15}
	cmpi.w	#0x7FFF, %d7		| pathological case check
	bne	fpd_round
	bra	fpd_gen_oflow
sub:	clr.l	%d3			| signs differed, so we subtract... /me groans
	sub.l	%d7, %d3
	subx.l	%d6, %d2
	subx.l	%d5, %d1
	bfffo	%d1{#0,#32}, %d6	| find MSB and test for zero result
	cmp.b	#32, %d6
	bne	subn1
	add.b	#64, %d6
	bra	subnx
subn1:	bfffo	%d2{#0,#32}, %d6
	cmp.b	#32, %d6
	bne	subn2
	add.b	#32, %d6
	bra	subnx
subn2:	bfffo	%d3{#0,#32}, %d6
	cmp.b	#32, %d6
	beq	fpd_gen_zero
subnx:	neg.w	%d6			| how many bits left will we shift mantissa?
	add.w	#95, %d6		| (enough to either normalize or, if that's
	bfextu	%d0{#16,#15}, %d7	| impossible, to set exponent = 0)
	cmp.w	%d6, %d7
	bge	snorm
	move.w	%d6, %d7
snorm:	bfextu	%d0{#16,#15}, %d6	| now (maybe) shift registers to normalize
	cmpi.b	#0x20, %d7		| (coarse-grained register swapping)
	ble	psbs
	move.l	%d2, %d1
	move.l	%d3, %d2
	clr.l	%d3
	sub.w	#0x20, %d6
	sub.w	#0x20, %d7
	cmpi.b	#0x20, %d7
	ble	psbs
	move.l	%d2, %d1
	clr.l	%d2
	sub.w	#0x20, %d6
	sub.w	#0x20, %d7
psbs:	beq	rtr
	move.w	%d7, %d4		| Post-Subtraction Bit Shifting...
	neg.w	%d4			| (fine-grained bit twaddling)
	add.w	#32, %d4
	lsl.l	%d7, %d1
	move.l	%d2, %d5
	lsr.l	%d4, %d5
	or.l	%d5, %d1
	lsl.l	%d7, %d2
	move.l	%d3, %d5
	lsr.l	%d4, %d5
	or.l	%d5, %d2
	lsl.l	%d7, %d3
	sub.w	%d7, %d6
rtr:	bfins	%d6, %d0{#16,#15}	| ready to round
	bra	fpd_round



|
|  fp_fsub: Perform an FSUB operation.  
|
|  This routine just performs some slightly different special-case checking
|  (the difference of two zeroes of opposing sign is always +0.0 in FSUB)
|  before jumping into the FADD code to add the negated source operand to
|  the destination operand.
|

fp_fsub:
	and.l	#0x00FF00FF, (fpe_fpsr,%a4)  | clear FPSR bytes that will be changed
	move.l	(%a0), %d0		| load source operand into %d0-%d2
	move.l	(#4,%a0), %d1
	move.l	(#8,%a0), %d2
	move.l	(%a1), %d4		| load destination operand into %d4-%d6
	move.l	(#4,%a1), %d5
	move.l	(#8,%a1), %d6
	jsr	fpd_handle_nan		| handle Not-A-Number inputs, if any
	bchg	#31, %d0		| negate source
	bfextu	%d0{#16,#15}, %d7	| is source an infinity?
	cmpi.w	#0x7FFF, %d7		
	blt	1f
	bfextu	%d4{#16,#15}, %d7	| source is INF... is dest an INF?
	cmpi.w	#0x7FFF, %d7
	bne	fpd_gen_inf		| if dest isn't INF, return inf w/sign of source
	xor.l	%d0, %d4		| so dest is INF... do the signs differ?
	btst	#31, %d4
	bne	fpd_gen_nan		| if so, return NAN
	bra	fpd_gen_inf		| if not, return infinity
1:	bfextu	%d4{#16,#15}, %d7	| is dest an infinity?
	cmpi.w	#0x7FFF, %d7
	blt	2f
	move.l	%d4, %d0
	bra	fpd_gen_inf		| yes... so leave it untouched (effectively)
2:	move.l	%d1, %d7		| is source a zero?
	or.l	%d2, %d7
	bne 	5f
	move.l	%d5, %d7		| source is zero. is dest zero of same sign?
	or.l	%d6, %d7
	beq	3f
	move.l	%d4, %d0
	move.l	%d5, %d1
	move.l	%d6, %d2
	bra	fpd_round		| dest isn't zero, so leave it be (effectively)
3:	move.l	%d0, %d7		| dest is zero... are the signs the same?
	xor.l	%d4, %d7
	bftst	#31, %d7
	bne	4f
	clr.l	%d0
	bra	fpd_gen_zero		| dest is zero of different sign... return +0.0
4:	clr.l	%d0			| signs match... see if rounding mode is RM
	bfextu	(fpe_fpcr,%a4){#4,#2}, %d7
	subq.l	#2, %d7			| cheaper than cmpi.b #2, %d7
	bne	1f			| if not RM mode, return +0.0
	bset	#31, %d0		| if RM mode, return -0.0
	bra	fpd_gen_zero
5:	move.l	%d5, %d7		| is dest a zero?
	or.l	%d6, %d7
	beq 	fpd_round		| if so, return source
	bra	unify










