|
|  fp_fmul: Perform an FMUL operation.
|

fp_fmul:
	|
	|  First, some preliminaries...
	|
	and.l	#0x00FF00FF, (fpe_fpsr,%a4)  | clear FPSR bytes that will be changed
	move.l	(%a0), %d0		| load source operand into %d0-%d2
	move.l	(#4,%a0), %d1
	move.l	(#8,%a0), %d2
	move.l	(%a1), %d4		| load destination operand into %d4-%d6
	move.l	(#4,%a1), %d5
	move.l	(#8,%a1), %d6
	jsr	fpd_handle_nan		| handle Not-A-Number inputs, if any

	|
	|  Compute sign - special-case results (except NAN's) always have a sign
	|  that is the exclusive OR of the two input signs (in other words, the
	|  "usual" sign).  Sign is stored in high bit of %d0.
	|
	btst	#31, %d4
	beq	1f
	bchg	#31, %d0

	|
	|  Special-case handling.  Please remain in your seats.
	|
1:	tst.l	%d1			| check for zero mantissa in source
	bne	chkDst			| (indicating either zero or infinity)
	tst.l	%d2
	bne	chkDst
	tst.l	%d5			| source has zero mantissa... see if dest does too
	bne	zmexit			| if not, just return source
	tst.l	%d6
	bne	zmexit
	bfextu	%d0{#16,#15}, %d7	| dest has zero mantissa... is source infinity?
	cmpi.w	#0x7FFF, %d7
	bne	idinf
	bfextu	%d4{#16,#15}, %d7	| source is infinity... is dest infinity too?
	cmpi.w	#0x7FFF, %d7
	beq	fpd_gen_inf		| if so, return an infinity (yay!)
	or.w	#0x2080, (fpe_fpsr+2,%a4)  | if not, set OPERR bit and return a NAN (boo!)
	bra	fpd_gen_nan
idinf:	bfextu	%d4{#16,#15}, %d7	| source is zero... is dest infinity?
	cmpi.w	#0x7FFF, %d7
	bne	fpd_gen_zero		| if not, return a zero (yay!)
	or.w	#0x2080, (fpe_fpsr+2,%a4)  | if so, set OPERR bit and return a NAN (boo!)
	bra	fpd_gen_nan
chkDst:	tst.l	%d5			| check for zero mantissa in dest
	bne	really			| (indicating either zero or infinity)
	tst.l	%d6			| if found, return dest
	bne	really
	bfins	%d4, %d0{#0,#31}	| yeah, it's kludgy
zmexit:	bfextu	%d0{#16,#15}, %d7	| return either zero or infinity...
	cmpi.w	#0x7FFF, %d7
	bne	fpd_gen_zero
	bra	fpd_gen_inf

	|
	|  Now really multiply.  We multiply the two 64-bit mantissas together
	|  to get a 128-bit result; note that the four registers containing this
	|  result are, from most to least significant, %d0, %d1, %d3, and %d2.
	|
really:	move.l	%d0, -(%sp)
	move.l	%d4, -(%sp)
	move.l	%d2, -(%sp)
	move.l	%d1, -(%sp)
	mulu.l	%d5, %d0-%d1
	mulu.l	%d6, %d3-%d2
	mulu.l	(%sp)+, %d4-%d5
	mulu.l	(%sp)+, %d7-%d6
	add.l	%d6, %d3
	addx.l	%d7, %d1
	bcc	1f			| cheaper than addx.l #0, %d0
	addq.l	#1, %d0	
1:	addx.l	%d5, %d3
	addx.l	%d4, %d1
	bcc	2f
	addq.l	#1, %d0	
2:	bfextu	(%sp)+{#16,#15}, %d6	| calculate exponent, store in %d7
	bfextu	(%sp){#16,#15}, %d7	| (leave old %d0 on the stack - it contains the sign)
	add.l	%d6, %d7
	sub.l	#0x3FFE, %d7

	|
	|  Normalize the gargantuan result.  We shift left until the explicit
	|  integer bit is set; afterwards, if the resulting number's exponent
	|  would be negative, we either denormalize the number if it's 
	|  salvageable, or otherwise generate an underflow.
	|
	btst	#31, %d0		| is result mantissa already normalized?
	bne	cof
	btst	#30, %d0		| no... can it be normalized in single bit shift?
	beq	hdnorm			| (usually the case, unless denorms are involved)
	lsl.l	#1, %d2
	roxl.l	#1, %d3
	roxl.l	#1, %d1
	roxl.l	#1, %d0
	subq.l	#1, %d7
	bra	cof
hdnorm:	bfffo	%d0{#0,#32}, %d6	| find MSB
	cmp.b	#32, %d6
	bne	muln1
	add.b	#96, %d6
	bra	mulnx
muln1:	bfffo	%d1{#0,#32}, %d6
	cmp.b	#32, %d6
	bne	muln2
	add.b	#64, %d6
	bra	mulnx
muln2:	bfffo	%d3{#0,#32}, %d6
	cmp.b	#32, %d6
	bne	muln3
	add.b	#32, %d6
	bra	mulnx
muln3:	bfffo	%d2{#0,#32}, %d6	| (no zero results with nonzero operands)
mulnx:	neg.w	%d6			| always normalize mantissa - deal with
	add.w	#127, %d6		| negative exponents later
	sub.l	%d6, %d7	
	cmpi.b	#32, %d6		| now (maybe) shift registers to normalize
	blt	pmbs			| (coarse-grained register swapping)
	move.l	%d1, %d0
	move.l	%d3, %d1
	move.l	%d2, %d3
	clr.l	%d2
	sub.b	#32, %d6
	cmpi.b	#32, %d6
	blt	pmbs
	move.l	%d1, %d0
	move.l	%d3, %d1
	clr.l	%d3
	sub.b	#32, %d6
	cmpi.b	#32, %d6
	blt	pmbs
	move.l	%d1, %d0
	clr.l	%d1
	sub.b	#32, %d6
pmbs:	tst.l	%d6			| need we go through with this?
	beq	cof
	move.w	%d6, %d5		| Post-Multiplication Bit Shifting...
	neg.w	%d5			| (fine-grained bit twaddling)
	add.w	#32, %d5
	lsl.l	%d6, %d0
	move.l	%d1, %d4
	lsr.l	%d5, %d4
	or.l	%d4, %d0
	lsl.l	%d6, %d1
	move.l	%d3, %d4
	lsr.l	%d5, %d4
	or.l	%d4, %d1
	lsl.l	%d6, %d3
	move.l	%d2, %d4
	lsr.l	%d5, %d4
	or.l	%d4, %d3
	lsl.l	%d6, %d2
cof:	tst.l	%d7			| is the exponent negative?
	bge	mietb			| if yes, might have to denormalize
	neg.l	%d7
	cmpi.w	#64, %d7		| yes... is result too small to be normalized?
	blt	1f
	move.l	(%sp)+, %d0		| (yep, it's too small)
	bra	fpd_gen_uflow
1:	cmpi.w	#32, %d7		| okay... is a shift of >=32 bits required?
	blt	pmbs2
	tst.l	%d2
	beq	mnorm1
	bset	#0, %d3
mnorm1:	move.l	%d3, %d2		| if so, move registers over (32 bit shift)
	move.l	%d1, %d3
	move.l	%d0, %d1
	clr.l	%d0
	sub.w	#32, %d7
pmbs2:	moveq.l	#32, %d6		| time to shift some serious bits.
	sub.l	%d7, %d6
	lsl.l	%d7, %d0
	move.l	%d1, %d5
	lsr.l	%d6, %d5
	or.l	%d5, %d0
	lsl.l	%d7, %d1
	move.l	%d3, %d5
	lsr.l	%d6, %d5
	or.l	%d5, %d1
	lsl.l	%d7, %d3
	move.l	%d2, %d5
	lsr.l	%d6, %d5
	or.l	%d5, %d3
	lsl.l	%d7, %d2
	beq	1f			| reduce mantissa to 96 bits prior to rounding
	bset	#0, %d3
1:	move	%d1, %d2
	move	%d0, %d1
	neg.l	%d7
mietb:	move.l	(%sp)+, %d0		| restore original %d0 (containing the sign)
	cmpi.w	#0x7FFF, %d7		| is exponent too big?
	bge	fpd_gen_oflow		| if so, overflow
	bfins	%d7, %d0{#16,#15}	| if not, punch out... it's been a long day.
	bra	fpd_round
