	.globl	fpd_handle_nan


|
|  fpd_handle_nan:  Unified Not-A-Number handling routine.
|
|  This routine Does The Right Thing(tm) if one or both inputs to an
|  operation are NAN's.  If such is the case, the routine sets the
|  appropriate status bits in FPSR, raises a SIGFPE if appropriate,
|  writes the proper NAN value to (%a1), and jumps to fp_end.  If
|  neither input is an NAN, the routine returns to the caller.
|  This routine is called at the start of most arithmetic opcodes.
|
|  Register usage:
|  %d0-%d2: source operand in extended format
|  %d4-%d6: destination operand in extended format
|  %d7:     mangled
|  %a1:     where to write the result, if appropriate; destination
|           operand is presumed to be there already
|  %a4:     pointer to global variables
|  others:  untouched
|

fpd_handle_nan:
	bfextu	%d4{#16,#15}, %d7	| is dest a NAN?
	cmpi.w	#0x7FFF, %d7
	blt	5f
	move.l	%d5, %d7
	or.l	%d6, %d7
	beq	5f
	move.b	#1, (fpe_fpsr, %a4)	| dest is NAN... set FPCC byte as sign-appropriate
	btst.l	#31, %d4
	beq	1f
	move.b	#9, (fpe_fpsr, %a4)
1:	btst.l	#30, %d5		| is dest an SNAN?
	bne	3f
	or.w	#0x4080, (fpe_fpsr+2, %a4)  | dest is SNAN - set SNAN bit in FPCR EXC byte
	btst.b	#6, (fpe_fpcr+2, %a4)	| is SNAN exception enabled (FPCR ENABLE byte)?
	beq	2f
	jsr	fpd_raise_sigfpe	| if so, throw exception
2:	bset.b	#6, (#4,%a1)		| convert dest to quiet NAN
	addq.l	#4, %sp			| pop calling address from stack...
	bra	fp_end			| ...and exit
3:	bfextu	%d0{#16,#15}, %d7	| is source an SNAN?
	cmpi.w	#0x7FFF, %d7
	blt	4f
	move.l	%d1, %d7
	or.l	%d2, %d7
	beq	4f
	btst.l	#30, %d1
	bne	4f
	or.w	#0x4080, (fpe_fpsr+2, %a4)  | source is SNAN - set SNAN bit in FPCR EXC byte
	btst.b	#6, (fpe_fpcr+2, %a4)	| is SNAN exception enabled (FPCR ENABLE byte)?
	beq	4f
	jsr	fpd_raise_sigfpe	| if so, throw exception
4:	addq.l	#4, %sp			| leave dest untouched
	bra	fp_end			| pop calling address and exit
5:	bfextu	%d0{#16,#15}, %d7	| is source a NAN?
	cmpi.w	#0x7FFF, %d7
	beq	6f
	rts
6:	move.l	%d1, %d7
	or.l	%d2, %d7
	bne	7f
	rts
7:	move.b	#1, (fpe_fpsr, %a4)	| source is NAN... set FPCC byte as sign-appropriate
	btst.l	#31, %d0
	beq	8f
	move.b	#9, (fpe_fpsr, %a4)
8:	btst.l	#30, %d1		| is source an SNAN?
	bne	10f
	or.w	#0x4080, (fpe_fpsr+2, %a4)  | source is SNAN - set SNAN bit in FPCR EXC byte
	btst.b	#6, (fpe_fpcr+2, %a4)	| is SNAN exception enabled (FPCR ENABLE byte)?
	beq	9f
	jsr	fpd_raise_sigfpe	| if so, throw exception
9:	bset.l	#30, %d1			| convert source to quiet NAN
10:	move.l	%d4, (%a1)		| write source->dest
	move.l	%d5, (#4, %a1)
	move.l	%d6, (#8, %a1)
	addq.l	#4, %sp			| pop calling address from stack...
	bra	fp_end			| ...and exit



|
|  fpd_gen_uflow:  Generate an underflow.
|
|  Depending on sign of the underflow, rounding mode, and rounding precision,
|  this routine writes either a zero or the appropriate smallest denormalized
|  number to (%a1); FPSR bits are also set appropriately.  When done, this
|  routine jumps to fp_end.
|
|  Register usage:
|  %d0:     high bit contains sign of underflow
|  %d3:     high bits are guard, round, sticky
|  %d4-%d5: mangled
|  %a1:     points to destination in memory
|  %a2:     mangled
|  %a4:     pointer to global variables
|  others:  unused
|

fpd_gen_uflow:
	lsr.l	#31, %d0
	bfextu	(fpe_fpcr,%a4){#4,#2}, %d4
	tst.b	%d4			| round-to-nearest mode?
	beq	1f			| if so, skip what follows
	cmpi.b	#1, %d4			| round-to-zero mode?
	beq	4f			| if so, return a naturally-signed zero
	move.l	%d4, %d5
	eor.b	%d0, %d5
	cmpi.b	#3, %d5			| directed round up?
	beq	2f			| if so, return smallest denorm
	cmpi.b	#2, %d5			| directed round down?
	beq	3f			| if so, return zero w/sign depending on mode
1:	bfextu	(fpe_fpcr,%a4){#6,#2}, %d4  | round-to-nearest...
	tst.b	%d4			| if not in extended precision,
	bne	4f			| and not (guard && (round || sticky)),
	btst.l	#31, %d3		| return naturally-signed zero
	bne	4f
	bftst	%d3{#0,#31}
	beq	4f
2:	bfextu	(fpe_fpcr,%a4){#6,#2}, %d4  | return smallest denormalized...
	mulu.w	#3, %d4
	lea	(9f:w,%pc,%d4*4), %a2	| look up value in table by precision
	move.l	(%a2)+, %d5
	ror.l	#1, %d0
	or.l	%d0, %d5
	move.l	%d5, (%a1)
	move.l	(%a2)+, (#4,%a1)
	move.l	(%a2), (#8,%a1)
	move.w	#0x0A28, %d5
	lsr.l	#4, %d0
	or.l	%d0, %d5
	or.l	%d5, (fpe_fpsr,%a4)	| set FPSR bits
	bra	fp_end
3:	move.b	%d4, %d0		| return zero with sign based on round mode
	and.b	#1, %d0			| (this just twiddles sign and falls through to:)
4:	ror.l	#1, %d0			| return zero with "natural" sign
	move.l	%d0, (%a1)
	clr.l	(#4,%a1)
	clr.l	(#8,%a1)
	move.l	#0x04000A28, %d5
	lsr.l	#4, %d0
	or.l	%d0, %d5
	or.l	%d5, (fpe_fpsr,%a4)	| set FPSR bits
	bra	fp_end
	|  table of smallest denormalized values:
	.align	4
9:	.long	0x00000000, 0x00000000, 0x00000001	| extended precision
	.long	0x3F6B0000, 0x80000000, 0x00000000	| single precision (2**-148)
	.long	0x3BCD0000, 0x80000000, 0x00000000	| double precision (2**-1074)
	.long	0x00000000, 0x00000000, 0x00000001	| undefined precision
	


|
|  fpd_gen_oflow: Generate an overflow.
|
|  Depending on sign of the overflow, rounding mode, and rounding precision,
|  this routine writes either an infinity or the largest value representable
|  in the current rounding precision to (%a1); FPSR bits are also set 
|  appropriately.  When done, this routine jumps to fp_end.
|
|  Register usage:
|  %d0:     high bit contains sign of overflow
|  %d1,%d4: mangled
|  %a1:     points to destination in memory
|  %a2:     mangled
|  %a4:     pointer to global variables
|  others:  unused
|

fpd_gen_oflow:
	lsr.l	#31, %d0
	bfextu	(fpe_fpcr,%a4){#4,#2}, %d4
	tst.b	%d4			| round-to-nearest mode?
	beq	1f			| if so, return infinity
	cmpi.b	#1, %d4			| round-to-zero?
	beq	2f			| if so, return largest value
	eor.b	%d0, %d4		| if ((RM && negative) || (RP && positive))
	cmpi.b	#3, %d4			|     return infinity
	bne	2f
1:	ror.l	#5, %d0			| return infinity...
	move.l	#0x02001248, %d1
	or.l	%d0, %d1
	or.l	%d1, (fpe_fpsr,%a4)
	lsl.l	#4, %d0
	bfset	%d0{#16,#15}
	move.l	%d0, (%a1)
	clr.l	(#4,%a1)
	clr.l	(#8,%a1)
	bra	fp_end
2:	bfextu	(fpe_fpcr,%a4){#6,#2}, %d4  | return largest value...
	mulu.w	#3, %d4
	lea	(9f:w,%pc,%d4*4), %a2	| look up value in table by precision
	move.l	(%a2)+, %d1
	ror.l	#1, %d0
	or.l	%d0, %d1
	move.l	%d1, (%a1)
	move.l	(%a2)+, (#4,%a1)
	move.l	(%a2), (#8,%a1)
	lsr.l	#4, %d0
	move.l	#0x02001248, %d1
	or.l	%d0, %d1
	or.l	%d1, (fpe_fpsr,%a4)
	bra	fp_end
	|  table of largest representable values:
	.align	4
9:	.long	0x7FFE0000, 0xFFFFFFFF, 0xFFFFFFFF	| extended precision
	.long	0x407E0000, 0xFFFFFF00, 0x00000000	| single precision
	.long	0x43FE0000, 0xFFFFFFFF, 0xFFFFF800	| double precision
	.long	0x7FFE0000, 0xFFFFFFFF, 0xFFFFFFFF	| undefined precision



|
|  fpd_gen_inf: Generate infinite result.
|
|  Stores an infinite result in (%a1), with sign as given in the high bit 
|  of %d0.  FPSR bits are set appropriately.  When done, this routine jumps
|  to fp_end.
|
|  Register usage:
|  %d0:     high bit contains sign of result
|  %d1:     mangled
|  %a1:     points to destination in memory
|  %a4:     pointer to global variables
|  others:  unused
|

fpd_gen_inf:
	moveq.b	#2, %d1
	btst	#31, %d0
	beq	1f
	moveq.b	#10, %d1
1f:	or.b	%d1,(fpe_fpsr,%a4)	| set FPSR bits
	bfset	%d0{#16:#15}
	move.l	%d0, (%a1)
	clr.l	(#4,%a1)
	clr.l	(#8,%a1)
	bra	fp_end



|
|  fpd_gen_zero: Generate zero result.
|
|  Stores a zero result in (%a1), with sign as given in the high bit of %d0.
|  FPSR bits are set appropriately.  When done, this routine jumps to fp_end.
|
|  Register usage:
|  %d0:     high bit contains sign of result
|  %d1:     mangled
|  %a1:     points to destination in memory
|  %a4:     pointer to global variables
|  others:  unused
|

fpd_gen_zero:
	moveq.b	#4, %d1
	btst	#31, %d0
	beq	1f
	moveq.b	#12, %d1
1f:	or.b	%d1,(fpe_fpsr,%a4)	| set FPSR bits
	bfclr	%d0{#0:#31}
	move.l	%d0, (%a1)
	clr.l	(#4,%a1)
	clr.l	(#8,%a1)
	bra	fp_end



|
|  fpd_gen_nan: Generate Not-A-Number result.
|
|  Stores a "quiet" +NAN result in (%a1).  FPSR bits are set appropriately.
|  When done, this routine jumps to fp_end.
|
|  Register usage:
|  %a1:     points to destination in memory
|  %a4:     pointer to global variables
|  others:  unused
|

fpd_gen_nan:
	ori.b 	#1, (fpe_fpsr,%a4)
	move.l	#0x7FFF0000, (%a1)
	clr.l	(#4,%a1)
	clr.l	(#8,%a1)
	bra	fp_end



|
|  fpd_raise_sigfpe:  Raise a floating-point exception.
|
|  This routine sends a SIGFPE to the current process, usually in lieu of an
|  exception that would otherwise be generated by the floating-point hardware.
|
|  The routine calls a C wrapper function that calls force_sig(SIGFPE, current).
|  We call a wrapper function rather than directly invoking force_sig because
|  "current" is a #define; we wish to avoid problems in case the definition of
|  current changes, or should conquering alien races decide to add SMP support 
|  to Linux/m68k far in the future.  Registers are saved to avoid any monkey 
|  business that may otherwise occur in the calling assembler routines.
|

fpd_raise_sigfpe:
	movem.l	%d0-%d7/%a0-%a6, -(%sp)
	jsr	SYMBOL_NAME(fpd_raise_sigfpe_c)
	movem.l	(%sp)+, %a0-%a6/%d0-%d7
	rts



|
|  fpd_norm_{dyadic,monadic}: Normalize inputs to an operation.
|
|  fpd_norm_dyadic normalizes the extended-precision operands in %d0-%d2 and
|  %d4-%d6; fpd_norm_monadic normalizes %d0-%d2 only.  It is assumed that no
|  operand is a zero, NaN, or infinity.  Called by many opcodes as a prelude
|  to performing an arithmetic operation.
|
|  Register usage:
|  %d0-%d2: monadic/first dyadic operand
|  %d4-%d6: second dyadic operand
|  %d3,%d7: mangled
|

fpd_norm_dyadic:
	btst.l	#31, %d5
	bne	fpd_norm_monadic
	bfffo	%d5{#0,#32},%d7
	cmpi.b	#32, %d7
	beq	1f
	add.b	#32, %d7
	bra	2f
1:	bfffo	%d6{#0,#32},%d7
2:	neg.w	%d7
	add.w	#63, %d7
	bfextu	%d4{#16,#15}, %d3
	cmp.w	%d3, %d7
	bge	3f
	move.l	%d3, %d7
3:	move.w	%d7, -(%sp)
	cmpi.b	#32, %d7
	blt	4f
	move.l	%d6, %d5
	clr.l	%d6
	sub.w	#32, %d7
	lsl.l	%d7, %d5
	bra	5f
4:	lsl.l	%d7, %d5
	move.l	%d6, %d3
	lsl.l	%d7, %d6
	neg.w	%d7
	add.w	#32, %d7
	lsr.l	%d7, %d3
	or.l	%d3, %d5
5:	bfextu	%d4{#16,#15}, %d3
	sub.w	(%sp)+, %d3
	bfins	%d3, %d4{#16,#15}
fpd_norm_monadic:
	btst.l	#31, %d1
	beq	6f
	rts
6:	bfffo	%d1{#0,#32},%d7
	cmpi.b	#32, %d7
	bne	7f
	add.b	#32, %d7
	bra	8f
7:	bfffo	%d2{#0,#32},%d7
8:	neg.w	%d7
	add.w	#63, %d7
	bfextu	%d0{#16,#15}, %d3
	cmp.w	%d3, %d7
	bge	9f
	move.l	%d3, %d7
9:	move.w	%d7, -(%sp)
	cmpi.b	#32, %d7
	blt	10f
	move.l	%d2, %d1
	clr.l	%d2
	sub.w	#32, %d7
	lsl.l	%d7, %d1
	bfextu	%d0{#16,#15}, %d3
	sub.w	(%sp)+, %d3
	bfins	%d3, %d0{#16,#15}
	rts
10:	lsl.l	%d7, %d1
	move.l	%d2, %d3
	lsl.l	%d7, %d2
	neg.w	%d7
	add.w	#32, %d7
	lsr.l	%d7, %d3
	or.l	%d3, %d1
	bfextu	%d0{#16,#15}, %d3
	sub.w	(%sp)+, %d3
	bfins	%d3, %d0{#16,#15}
	rts



|
|  fpd_round: Round a result based on current rounding mode and precision.
|
|  Rounds the number in %d0-%d3, and writes the rounded result to (%a1);
|  falls through to fpd_judgement when finished.  The input is similar to 
|  an extended-format number: sign and exponent in the high word of %d0, 
|  with 96 bits of mantissa (including explicit integer bit) in %d1-%d3.
|  It is assumed that this input has already been normalized, and is not 
|  a NaN or infinity.
|
|  Register usage:
|  %d0-%d3: input
|  %d4-%d7: mangled
|  %a1:     points to destination in memory
|  %a4:     pointer to global variables
|  others:  unused
|

fpd_round:
	bfextu	(fpe_fpcr,%a4){#6,#2}, %d6  | determine rounding precision, truncate to fit
	cmpi.b	#1, %d6			| single precision?
	beq	2f
	cmpi.b	#2, %d6			| double precision?
	bne	5f
	moveq.l	#11, %d4		| double precision - shift 11 bits right
	moveq.l	#21, %d5
	bftst	%d3{#0,#11}
	bne	1f
	bset	%d4, %d3
1:	lsr.l	%d4, %d3
	move.l	%d2, %d7
	lsl.l	%d5, %d7
	or.l	%d7, %d3
	lsr.l	%d4, %d2
	move.l	%d1, %d7
	lsl.l	%d5, %d7
	or.l	%d7, %d2
	lsr.l	%d4, %d1
	bra	5f
2:	moveq.l	#24, %d5		| single precision - shift 40 bits right
	tst.w	%d3
	bne	3f
	bftst	%d2{#0,#24}
	beq	4f
3:	bset	%d5, %d2
4:	lsr.l	%d5, %d2
	move.l	%d1, %d3
	lsl.l	#8, %d3
	or.l	%d2, %d3
	move.l	%d1, %d2
	lsr.l	%d5, %d2
	clr.l	%d1
5:	bfextu	(fpe_fpcr,%a4){#4,#2}, %d4  | determine rounding mode
	tst.b	%d4			| round-to-nearest mode? if so,
	beq	6f			| round up on (guard && (LSB || round || sticky))
	cmpi.b	#1, %d4			| round-to-zero mode?
	beq	8f			| if so, truncate the number
	move.l	%d0, %d5
	lsr.l	#31, %d5
	eor.b	%d4, %d5
	cmpi.b	#3, %d5			| directed round up?
	bne	8f			| if not, (directed down), truncate the number
	tst.l	%d3			| if so, round up on (guard || round || sticky)
	bne	7f
	bra	8f
6:	btst	#31, %d3	
	beq	8f
	bftst	%d3{#0,#31}
	bne	7f
	btst	#0, %d2
	beq	8f
7:	addq.l	#1, %d2
	bcc	8f
	addq.l	#1, %d1
	bcc	8f
	roxr.l	#1, %d1
	roxr.l	#1, %d2
	bfextu	%d0{#16,#15}, %d5
	addq.w	#1, %d5
	cmpi.w	#0x7FFF, %d5
	bge	fpd_gen_oflow
	bfins	%d5, %d0{#16,#15}
8:	tst.l	%d3			| inexact result?
	beq	9f
	or.w	#0x0204, (fpe_fpsr+2, %a4)  | if so, set INEX2 bit in FPSR
9:	cmpi.b	#1, %d6			| time to truncate... single precision mode?
	beq	11f
	cmpi.b	#2, %d6			| double precision mode?
	bne	13f
	moveq.l	#21, %d5		| truncating in double precision mode...
	bfextu	%d0{#16,#15}, %d4
	btst	%d5, %d1		| check for carryover
	bne	10f
	lsr.l	#1, %d1
	roxr.l	#1, %d2
	addq.w	#1, %d4
10:	cmpi.w	#0x43FE, %d4		| make sure exponent is within bounds
	bgt	fpd_gen_oflow
	cmpi.w	#0x3BCD, %d4
	blt	fpd_gen_uflow
	bfins	%d4, %d0{#16,#15}
	moveq.l	#11, %d4		| shift mantissa 11 bits left
	lsl.l	%d4, %d1
	move.l	%d2, %d3
	lsr.l	%d5, %d3
	or.l	%d3, %d1
	lsl.l	%d4, %d2
	bra	13f

11:	bfextu	%d0{#16,#15}, %d4	| truncating in single precision mode...
	btst	#24, %d2		| check for carryover
	bne	12f
	lsr.l	#1, %d2
	addq.w	#1, %d4
12:	cmpi.w	#0x407E, %d4		| make sure exponent is within bounds
	bgt	fpd_gen_oflow
	cmpi.w	#0x3F6B, %d4
	blt	fpd_gen_uflow
	bfins	%d4, %d0{#16,#15}
	move.l	%d2, %d1		| shift mantissa 40 bits left
	clr.l	%d2
	lsl.l	#8, %d1
13:	move.l	%d0, (%a1)
	move.l	%d1, (#4,%a1)
	move.l	%d2, (#8,%a1)
	| fall through to...



|
|  fpd_judgement: Set the FPSR bits according to input value.
|
|  Sets bits in the FPSR condition code byte according to the operand in
|  %d0-%d2; jumps to fp_end when finished.  It is assumed that the operand
|  is not an infinity or NaN, and that the FPSR has been cleared.
|
|  Register usage:
|  %d0-%d3: input
|  %a4:     pointer to global variables
|  others:  unused
|

fpd_judgement:
	btst	#31, %d0
	beq	2f
	tst.l	%d1
	bne	1f
	tst.l	%d2
	bne	1f
	move.b	#8, (fpe_fpsr, %a4)
	bra	fp_end
1:	move.b	#12, (fpe_fpsr, %a4)
	bra	fp_end
2:	tst.l	%d1
	bne	3f
	tst.l	%d2
	bne	3f
	move.b	#4, (fpe_fpsr, %a4)
3:	bra	fp_end
