#include "fp.h"

	.globl	fp_scan

	.data
fp_temp:
	.space	12

| %d6 - first two instr words
| %d5 - operand size
| %d4 - further instr word

	.text
fp_scan:
| normal fpu instruction? (this excludes fsave/frestore)
	moves_b	(%a5),%d0
	cmp.b	#0xfc,%d0
	jne	fp_nonstd
| first two instruction words are kept in %d6
	moves_l	(%a5)+,%d6		
	bfextu	%d6{#8,#2},%d0		| test fpu instruction type
	jmp	([0f:w,%pc,%d0*4])

	.align	4
0:
	.long	fp_generic, fp_fpscc, fp_fpbccw, fp_fpbccl

fp_generic:
	P("fpgeneric")
	bfextu	%d6{#16,#3},%d0		| test fpu generic type 
	jmp	([0f:w,%pc,%d0*4])

	.align	4
0:
	.long	fp_gen, fp_ill, fp_gen, fp_fmove_fp2mem
	.long	fp_fmovem_mem2cr, fp_fmovem_cr2mem
	.long	fp_fmovem_fp, fp_fmovem_fp

| these are more or less all arithmetic instructions 

fp_gen:
	P("fp_gen")
	btst	#14,%d6
	jne	fp_fromcpu

| source is a fpu register

	bfextu	%d6{#19,#3},%d0
	P1("source fp%d", %d0)
	jsr	fp_get_fpu_reg
	jra	fp_getdest

| source is cpu register or somewhere in memory

fp_fromcpu:
	/* sort out fmovecr, keep data size in %d5 */
	bfextu	%d6{#19,#3},%d5
	jmp	([0f:w,%pc,%d5*4])

	.align	4
0:
#ifdef DEBUG
	.long	fp_format_long, fp_format_single
	.long	fp_format_extended, fp_format_packed
	.long	fp_format_word, fp_format_double
	.long	fp_format_byte, fp_fmovecr

fp_format_long:
	P("source format: long")
	jra	fp_getsource

fp_format_single:
	P("source format: single")
	jra	fp_getsource

fp_format_extended:
	P("source format: extended")
	jra	fp_getsource

fp_format_packed:
	P("source format: packed")
	jra	fp_getsource

fp_format_word:
	P("source format: word")
	jra	fp_getsource

fp_format_double:
	P("source format: double")	
	jra	fp_getsource

fp_format_byte:
	P("source format: byte")
	jra	fp_getsource
#else
	.long	fp_getsource, fp_getsource
	.long	fp_getsource, fp_getsource
	.long	fp_getsource, fp_getsource
	.long	fp_getsource, fp_fmovecr
#endif

fp_fmovecr:
	P("fp_fmovecr")
	bfextu	%d6{#27,#5},%d0
	move.l	%d0,%d1
	add.l	%d0,%d0
	add.l	%d1,%d0
	lea	(fp_constants,%d0*4),%a0
	move.l	#0x801ec0ff,%d0
	addq.l	#1,%d1
	lsl.l	%d1,%d0
	jcc	1f
	bset	#1,fp_sr+2
	P("inexact result")
1:	moveq	#-128,%d0				| continue with fmove
	and.l	%d0,%d6
	jra	fp_getdest

	.data

	.align	4
fp_constants:
	.long	0x40000000,0xc90fdaa2,0x2168c235	| pi
	.extend	0,0,0,0,0,0,0,0,0,0
	.long	0x3ffd0000,0x9a209a84,0xfbcff798	| log10(2)
	.long	0x40000000,0xadf85458,0xa2bb4a9a	| e
	.long	0x3fff0000,0xb8aa3b29,0x5c17f0bc	| log2(e)
	.long	0x3ffd0000,0xde5bd8a9,0x37287195	| log10(e)
	.long	0x00000000,0x00000000,0x00000000	| 0.0
	.long	0x3ffe0000,0xb17217f7,0xd1cf79ac	| 1n(2)
	.long	0x40000000,0x935d8ddd,0xaaa8ac17	| 1n(10)
	.long	0x3fff0000,0x80000000,0x00000000	| 10^0
	.long	0x40020000,0xa0000000,0x00000000	| 10^1
	.long	0x40050000,0xc8000000,0x00000000	| 10^2
	.long	0x400c0000,0x9c400000,0x00000000	| 10^4
	.long	0x40190000,0xbebc2000,0x00000000	| 10^8
	.long	0x40340000,0x8e1bc9bf,0x04000000	| 10^16
	.long	0x40690000,0x9dc5ada8,0x2b70b59e	| 10^32
	.long	0x40d30000,0xc2781f49,0xffcfa6d5	| 10^64
	.long	0x41a80000,0x93ba47c9,0x80e98ce0	| 10^128
	.long	0x43510000,0xaa7eebfb,0x9df9de8e	| 10^256
	.long	0x46a30000,0xe319a0ae,0xa60e91c7	| 10^512
	.long	0x4d480000,0xc9767586,0x81750c17	| 10^1024
	.long	0x5a920000,0x9e8b3b5d,0xc53d5de5	| 10^2048
	.long	0x75250000,0xc4605202,0x8a20979b	| 10^4096

	.text

fp_getsource:
	| encode addressing mode for source
	encode_addrmode(%d0)
	jmp	([0f:w,%pc,%d0*4])

	.align	4
0:
	.long 	fp_data, fp_ill
	.long 	fp_indirect, fp_postinc
	.long 	fp_predecr, fp_disp16
	.long 	fp_extmode0, fp_extmode1

	| addressing mode: data register direct
fp_data:
	encode_addrreg(%d0)
	jsr	fp_get_data_reg
	P1("source d%d", %d0)
	lea	(fp_temp:w,%pc),%a0
	jmp	([0f:w,%pc,%d5*4])

	.align	4
0:
	.long	fp_data_long, fp_data_single
	.long	fp_ill, fp_ill
	.long	fp_data_word, fp_ill
	.long	fp_data_byte, fp_ill

fp_data_byte:
	extb.l	%d0
	jra	fp_data_long

fp_data_word:
	ext.l	%d0

fp_data_long:
	jsr	fp_conv_long2ext
	jra	fp_getdest

fp_data_single:
	jsr	fp_conv_single2ext
	jra	fp_getdest

	| addressing mode: address register indirect
fp_indirect:
	encode_addrreg(%d0)
	P1("source (a%d)", %d0)
	jsr	fp_get_addr_reg
	jra	fp_fetchsource

	| addressing mode: address register indirect with postincrement
fp_postinc:
	encode_addrreg(%d0)
	test_byte_sp_transf(%d0, %d1)
	P1("source (a%d)+", %d0)
	jsr	fp_get_addr_reg
	do_post_inc(%d1, %a1)
	jra	fp_fetchsource

	| addressing mode: address register indirect with predecrement
fp_predecr:
	encode_addrreg(%d0)
	test_byte_sp_transf(%d0, %d1)
	P1("source -(a%d)", %d0)
	jsr	fp_get_addr_reg
	do_pre_decr(%d1)
	jra	fp_fetchsource

fp_disp16:
	test_basereg_d16()
	jeq	1f
#ifdef DEBUG
	moves_w	(%a5),%a0
	P1("base (%x,pc)",%a0)
#endif
	jsr	fp_get_pc
	jra	2f
1:	encode_addrreg(%d0)
#ifdef DEBUG
	moves_w	(%a5),%a0
	P2("source (%x,a%d)", %a0, %d0)
#endif
	jsr	fp_get_addr_reg
2:	do_disp16(%a1, %a0)
	jra	fp_fetchsource

fp_extmode0:
	get_extword()
	test_exttype()
	jne	fp_fullext

	test_basereg_ext()
	jeq	1f
	P("base pc")
	jsr	fp_get_pc
	jra	2f
1:	encode_addrreg(%d0)
	P1("base a%d", %d0)
	jsr	fp_get_addr_reg
2:	move.l	%a0,%a1
	encode_index(%d1, %d0)
	add.l	%d0,%a1
	do_disp8(%a1, %a0, %d0)
	jra	fp_getdest

fp_fullext:
	sub.l	%a1,%a1
	test_suppr_base()		| base register suppressed?
	jne	3f
	test_basereg_ext()
	jeq	1f
	P("base pc")
	jsr	fp_get_pc
	jra	2f
1:	encode_addrreg(%d0)
	P1("base a%d", %d0)
	jsr	fp_get_addr_reg
2:	move.l	%a0,%a1
#ifdef DEBUG
	jra	4f
3:	P("base register suppresed")
4:
#else
3:
#endif
	encode_basesize(%d0)
	jmp	([0f:w,%pc,%d0*4])

	.align	4
0:
#ifdef DEBUG
	.long	fp_ill, fp_base0
	.long	fp_base16, fp_base32

fp_base0:
	P("no disp")
	jra	fp_outer
#else
	.long	fp_ill, fp_outer
	.long	fp_base16, fp_base32
#endif

fp_base16:
	get_base16(%a0)
	jra	1f

fp_base32:
	get_base32(%a0)
1:	add.l	%a0,%a1

fp_outer:
	do_preindex(%a1, %d0)
	test_suppr_index()		| index register suppressed?
	jne	fp_postindex
	encode_index(%d1, %d0)
	add.l	%d0,%a1

fp_postindex:
	do_postindex(%a1)

	encode_outersize(%d0)
	jmp	([0f:w,%pc,%d0*4])

	.align	4
0:
#ifdef DEBUG
	.long	fp_fullext_fin, fp_outer0
	.long	fp_outer16, fp_outer32

fp_outer0:
	P("no odisp")
	jra	fp_fullext_fin
#else
	.long	fp_fullext_fin, fp_fullext_fin
	.long	fp_outer16, fp_outer32
#endif

fp_outer16:
	get_outer16(%a0)
	jra	1f

fp_outer32:
	get_outer32(%a0)
1:	add.l	%a1,%a0
	jra	fp_fetchsource

fp_fullext_fin:
	move.l	%a1,%a0
	jra	fp_fetchsource

fp_extmode1:
	encode_addrreg(%d0)
	jmp	([0f:w,%pc,%d0*4])

	.align	4
0:
	.long	fp_abs_short, fp_abs_long
	.long	fp_disp16, fp_extmode0
	.long	fp_immediate, fp_ill
	.long	fp_ill, fp_ill

fp_abs_short:
	get_abs_short(%a0)
	jra	fp_fetchsource

fp_abs_long:
	get_abs_long(%a0)
	jra	fp_fetchsource

fp_immediate:
	P("#x")
	move.l	%a5,%a0
	move.w	(fp_datasize,%d5*2),%d0
	btst	#0,%d0
	jeq	1f
	addq.w	#1,%d0
	addq.l	#1,%a0
1:	add.w	%d0,%a5
	jra	fp_fetchsource

fp_fetchsource:
	jmp	([0f:w,%pc,%d5*4])

	.align	4
0:	.long	fp_long, fp_single
	.long	fp_ext, fp_pack
	.long	fp_word, fp_double
	.long	fp_byte, fp_ill

fp_long:
	moves_l	(%a0),%d0
	lea	(fp_temp:w,%pc),%a0
	jsr	fp_conv_long2ext
	jra	fp_getdest

fp_single:
	moves_l	(%a0),%d0
	lea	(fp_temp:w,%pc),%a0
	jsr	fp_conv_single2ext
	jra	fp_getdest

fp_ext:
	move.l	%a0,%a1
	lea	(fp_temp:w,%pc),%a0
	jsr	fp_conv_ext2ext
	jra	fp_getdest

fp_pack:
	/* not supported yet */
	jra	fp_ill

fp_word:
	moves_w	(%a0),%d0
	ext.l	%d0
	lea	(fp_temp:w,%pc),%a0
	jsr	fp_conv_long2ext
	jra	fp_getdest

fp_double:
	move.l	%a0,%a1
	lea	(fp_temp:w,%pc),%a0
	jsr	fp_conv_double2ext
	jra	fp_getdest

fp_byte:
	moves_b	(%a0),%d0
	extb.l	%d0
	lea	(fp_temp:w,%pc),%a0
	jsr	fp_conv_long2ext
	jra	fp_getdest

fp_getdest:
|	P4("s at %p: %08x %08x %08x", %a0, (%a0), (4,%a0), (8,%a0))
	move.l	%a0,%a1
	bfextu	%d6{#22,#3},%d0
|	P1("dest fp%d", %d0)
	jsr	fp_get_fpu_reg
|	P4("d at %p: %08x %08x %08x", %a0, (%a0), (4,%a0), (8,%a0))
	bfextu	%d6{#25,#7},%d0
	jmp	([0f:w,%pc,%d0*4])

	.align	4
0:
	.long	fp_fmove_mem2fp, fp_fint, fp_fsinh, fp_fintrz
	.long	fp_fsqrt, fp_ill, fp_flognp1, fp_ill
	.long	fp_fetoxm1, fp_ftanh, fp_fatan, fp_ill
	.long	fp_fasin, fp_fatanh, fp_fsin, fp_ftan
	.long	fp_fetox, fp_ftwotox, fp_ftentox, fp_ill
	.long	fp_flogn, fp_flog10, fp_flog2, fp_ill
	.long	fp_fabs, fp_fcosh, fp_fneg, fp_ill
	.long	fp_facos, fp_fcos, fp_fgetexp, fp_fgetman
	.long	fp_fdiv, fp_fmod, fp_fadd, fp_fmul
	.long	fp_fsgldiv, fp_frem, fp_fscale, fp_sglmul
	.long	fp_fsub, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_fsincos0, fp_fsincos1, fp_fsincos2, fp_fsincos3
	.long	fp_fsincos4, fp_fsincos5, fp_fsincos6, fp_fsincos7
	.long	fp_fcmp, fp_ill, fp_ftst, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_fsmove, fp_fssqrt, fp_ill, fp_ill
	.long	fp_fdmove, fp_fdsqrt, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_fsabs, fp_ill, fp_fsneg, fp_ill
	.long	fp_fdabs, fp_ill, fp_fdneg, fp_ill
	.long	fp_fsdiv, fp_ill, fp_fsadd, fp_fsmul
	.long	fp_fddiv, fp_ill, fp_fdadd, fp_fdmul
	.long	fp_fssub, fp_ill, fp_ill, fp_ill
	.long	fp_fdsub, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill
	.long	fp_ill, fp_ill, fp_ill, fp_ill

fp_fmove_mem2fp:
|	P("fp_fmove_mem2fp")
	move.l	(%a1)+,(%a0)+
	move.l	(%a1)+,(%a0)+
	move.l	(%a1),(%a0)
	subq.l	#8,%a0
|	P4("d at %p: %08x %08x %08x", %a0, (%a0), (4,%a0), (8,%a0))
	jra	fp_finish

fp_fint:
	P("fp_fint")
	jra	fp_end

fp_fsinh:
	P("fp_fsinh")
	jra	fp_end

fp_fintrz:
	P("fp_fintrz")
	jra	fp_end

fp_fsqrt:
	P("fp_fsqrt")
	jra	fp_end

fp_flognp1:
	P("fp_flognp1")
	jra	fp_end

fp_fetoxm1:
	P("fp_fetoxm1")
	jra	fp_end

fp_ftanh:
	P("fp_ftanh")
	jra	fp_end

fp_fatan:
	P("fp_fatan")
	jra	fp_end

fp_fasin:
	P("fp_fasin")
	jra	fp_end

fp_fatanh:
	P("fp_fatanh")
	jra	fp_end

fp_fsin:
	P("fp_fsin")
	jra	fp_end

fp_ftan:
	P("fp_ftan")
	jra	fp_end

fp_fetox:
	P("fp_fetox")
	jra	fp_end

fp_ftwotox:
	P("fp_ftwotox")
	jra	fp_end

fp_ftentox:
	P("fp_ftentox")
	jra	fp_end

fp_flogn:
	P("fp_flogn")
	jra	fp_end

fp_flog10:
	P("fp_flog10")
	jra	fp_end

fp_flog2:
	P("fp_flog2")
	jra	fp_end

fp_fabs:
	P("fp_fabs")
	jra	fp_end

fp_fcosh:
	P("fp_fcosh")
	jra	fp_end

fp_fneg:
	P("fp_fneg")
	jra	fp_end

fp_facos:
	P("fp_facos")
	jra	fp_end

fp_fcos:
	P("fp_fcos")
	jra	fp_end

fp_fgetexp:
	P("fp_fgetexp")
	jra	fp_end

fp_fgetman:
	P("fp_fgetman")
	jra	fp_end

fp_fdiv:
	P("fp_fdiv")
	jra	fp_end

fp_fmod:
	P("fp_fmod")
	jra	fp_end

fp_fadd:
	P("fp_fadd")
	jra	fp_end

fp_fmul:
	P("fp_fmul")
	jra	fp_end

fp_fsgldiv:
	P("fp_fsgldiv")
	jra	fp_end

fp_frem:
	P("fp_frem")
	jra	fp_end

fp_fscale:
	P("fp_fscale")
	jra	fp_end

fp_sglmul:
	P("fp_sglmul")
	jra	fp_end

fp_fsub:
	P("fp_fsub")
	jra	fp_end

fp_fsincos0:
	P("fp_fsincos0")
	jra	fp_end

fp_fsincos1:
	P("fp_fsincos1")
	jra	fp_end

fp_fsincos2:
	P("fp_fsincos2")
	jra	fp_end

fp_fsincos3:
	P("fp_fsincos3")
	jra	fp_end

fp_fsincos4:
	P("fp_fsincos4")
	jra	fp_end

fp_fsincos5:
	P("fp_fsincos5")
	jra	fp_end

fp_fsincos6:
	P("fp_fsincos6")
	jra	fp_end

fp_fsincos7:
	P("fp_fsincos7")
	jra	fp_end

fp_fcmp:
	P("fp_fcmp")
	jra	fp_end

fp_ftst:
	P("fp_ftst")
	jra	fp_end

fp_fsmove:
	P("fp_fsmove")
	jra	fp_end

fp_fssqrt:
	P("fp_fssqrt")
	jra	fp_end

fp_fdmove:
	P("fp_fdmove")
	jra	fp_end

fp_fdsqrt:
	P("fp_fdsqrt")
	jra	fp_end

fp_fsabs:
	P("fp_fsabs")
	jra	fp_end

fp_fsneg:
	P("fp_fsneg")
	jra	fp_end

fp_fdabs:
	P("fp_fdabs")
	jra	fp_end

fp_fdneg:
	P("fp_fdneg")
	jra	fp_end

fp_fsdiv:
	P("fp_fsdiv")
	jra	fp_end

fp_fsadd:
	P("fp_fsadd")
	jra	fp_end

fp_fsmul:
	P("fp_fsmul")
	jra	fp_end

fp_fddiv:
	P("fp_fddiv")
	jra	fp_end

fp_fdadd:
	P("fp_fdadd")
	jra	fp_end

fp_fdmul:
	P("fp_fdmul")
	jra	fp_end

fp_fssub:
	P("fp_fssub")
	jra	fp_end

fp_fdsub:
	P("fp_fdsub")
	jra	fp_end

fp_fmove_fp2mem:
	P("fp_fmove_fp2mem")
	jra	fp_end

fp_fmovem_mem2cr:
	P("fp_fmovem_mem2cr")
	jra	fp_end

fp_fmovem_cr2mem:
	P("fp_fmovem_cr2mem")
	jra	fp_end

fp_nonstd:
	P("nonstd")
	addq.l	#2,%a5
	jra	fp_end


	.data
	.align	4

fp_datasize:
	.word	4, 4, 12, 20, 2, 8, 1, 0
